# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info


def coco_names_80(label: int) -> str:
    """Return COCO category name (80 classes)

    Args:
        label (int): Category ID

    Returns:
        name (str): Category name
    """

    coco_dict = {
        1: "person",
        2: "bicycle",
        3: "car",
        4: "motorcycle",
        5: "airplane",
        6: "bus",
        7: "train",
        8: "truck",
        9: "boat",
        10: "traffic light",
        11: "fire hydrant",
        12: "stop sign",
        13: "parking meter",
        14: "bench",
        15: "bird",
        16: "cat",
        17: "dog",
        18: "horse",
        19: "sheep",
        20: "cow",
        21: "elephant",
        22: "bear",
        23: "zebra",
        24: "giraffe",
        25: "backpack",
        26: "umbrella",
        27: "handbag",
        28: "tie",
        29: "suitcase",
        30: "frisbee",
        31: "skis",
        32: "snowboard",
        33: "sports ball",
        34: "kite",
        35: "baseball bat",
        36: "baseball glove",
        37: "skateboard",
        38: "surfboard",
        39: "tennis racket",
        40: "bottle",
        41: "wine glass",
        42: "cup",
        43: "fork",
        44: "knife",
        45: "spoon",
        46: "bowl",
        47: "banana",
        48: "apple",
        49: "sandwich",
        50: "orange",
        51: "broccoli",
        52: "carrot",
        53: "hot dog",
        54: "pizza",
        55: "donut",
        56: "cake",
        57: "chair",
        58: "couch",
        59: "potted plant",
        60: "bed",
        61: "dining table",
        62: "toilet",
        63: "tv",
        64: "laptop",
        65: "mouse",
        66: "remote",
        67: "keyboard",
        68: "cell phone",
        69: "microwave",
        70: "oven",
        71: "toaster",
        72: "sink",
        73: "refrigerator",
        74: "book",
        75: "clock",
        76: "vase",
        77: "scissors",
        78: "teddy bear",
        79: "hair drier",
        80: "toothbrush",
    }

    return coco_dict[int(label)]


def coco_names_91(label: int) -> str:
    """Return COCO category name (91 classes)

    Args:
        label (int): Category ID

    Returns:
        name (str): Category name
    """

    coco_dict = {
        1: "person",
        2: "bicycle",
        3: "car",
        4: "motorbike",
        5: "aeroplane",
        6: "bus",
        7: "train",
        8: "truck",
        9: "boat",
        10: "trafficlight",
        11: "firehydrant",
        12: "streetsign",
        13: "stopsign",
        14: "parkingmeter",
        15: "bench",
        16: "bird",
        17: "cat",
        18: "dog",
        19: "horse",
        20: "sheep",
        21: "cow",
        22: "elephant",
        23: "bear",
        24: "zebra",
        25: "giraffe",
        26: "hat",
        27: "backpack",
        28: "umbrella",
        29: "shoe",
        30: "eyeglasses",
        31: "handbag",
        32: "tie",
        33: "suitcase",
        34: "frisbee",
        35: "skis",
        36: "snowboard",
        37: "sportsball",
        38: "kite",
        39: "baseballbat",
        40: "baseballglove",
        41: "skateboard",
        42: "surfboard",
        43: "tennisracket",
        44: "bottle",
        45: "plate",
        46: "wineglass",
        47: "cup",
        48: "fork",
        49: "knife",
        50: "spoon",
        51: "bowl",
        52: "banana",
        53: "apple",
        54: "sandwich",
        55: "orange",
        56: "broccoli",
        57: "carrot",
        58: "hotdog",
        59: "pizza",
        60: "donut",
        61: "cake",
        62: "chair",
        63: "sofa",
        64: "pottedplant",
        65: "bed",
        66: "mirror",
        67: "diningtable",
        68: "window",
        69: "desk",
        70: "toilet",
        71: "door",
        72: "tvmonitor",
        73: "laptop",
        74: "mouse",
        75: "remote",
        76: "keyboard",
        77: "cellphone",
        78: "microwave",
        79: "oven",
        80: "toaster",
        81: "sink",
        82: "refrigerator",
        83: "blender",
        84: "book",
        85: "clock",
        86: "vase",
        87: "scissors",
        88: "teddybear",
        89: "hairdrier",
        90: "toothbrush",
        91: "hairbrush",
    }

    return coco_dict[int(label)]


def voc_names(label: int) -> str:
    """Return VOC category name (20 classes)

    Args:
        label (int): Category ID

    Returns:
        name (str): Category name
    """

    voc_dict = {
        1: "aeroplane",
        2: "bicycle",
        3: "bird",
        4: "boat",
        5: "bottle",
        6: "bus",
        7: "car",
        8: "cat",
        9: "chair",
        10: "cow",
        11: "dining table",
        12: "dog",
        13: "horse",
        14: "motorbike",
        15: "person",
        16: "potted plant",
        17: "sheep",
        18: "sofa",
        19: "train",
        20: "tv / monitor",
    }

    return voc_dict[int(label)]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

__version__ = "0.2.1"



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import time
from math import isnan
from pathlib import Path

from PIL import Image
from pycocotools import mask as mask_api
from tqdm import tqdm

from pixano import transforms
from pixano.core import arrow_types

from .data_loader import DataLoader


class PixanoLoader(DataLoader):
    """Pixano Data Loader

    Attributes:
        dataset (dict): Dataset
        info (dict): Dataset info
        iter_data (iter): Data iterable
    """

    def __init__(
        self, workspace: str, ann_file: str, img_path: Path = None, view: str = "image"
    ):
        """Initalize COCO Data Loader

        Args:
            workspace (str): Data path
            ann_file (str): Annotation file name
            img_path (Path, optional): Image path. Defaults to None.
            view (str, optional): Image view name. Defaults to "image".
        """

        if ann_file is None:
            print(
                "PixanoLoader: Please provide an annotation file, or use another Loader if no annotations"
            )
            return
        self.info = {}

        # load dataset
        self.dataset = dict()
        print("loading annotations into memory...")
        tic = time.time()
        workspace = Path(workspace)
        with open(workspace / ann_file, "r") as f:
            raw_json = json.load(f)
        assert type(raw_json) == dict, "annotation file format {} not supported".format(
            type(raw_json)
        )
        print("Done (t={:0.2f}s)".format(time.time() - tic))

        # create index
        print("creating index...")
        tic = time.time()
        self.dataset = self.getFeaturesFromJSON(raw_json, workspace, img_path, view)
        self.iter_data = iter(self.dataset)
        print("Done (t={:0.2f}s)".format(time.time() - tic))
        self.info["nb_images"] = len(self.dataset)
        self.info["task"] = raw_json["task_name"]
        self.info["data_type"] = raw_json["data"]["type"]

    # Pixano wrapper
    # Pixano json has "annotations", a list of all annotations (image_id inside)
    # and "data", list of image path and id (timestamp)
    # (path is relative to workspace, and may be wrong: we keep only basename)
    # contains also Pixano task name
    """ For Reference: Pixano "annotations" structure
    class PixanoAnns:
    {
        "annotations": [
            'associated_to': str
            'category': str
            'geometry': {
                'mvertices': list[int]
                'type': str
                'vertices': list[int]
            }
            'id': str
            'options': {
                'occlusion': float
                'truncation': float
            }
            'timestamp': int
        ]
        "data": {
            "children": [
                {
                    "path": "images/cylindric/test/20170320_163113/cam_0/20170320_163113_cam_0_00006300.jpg",
                    "timestamp": 0
                }]
            "path": "images/cylindric/test/20170320_163113/cam_0",
            "type": "sequence_image"
        },
        "task_name": "object2d"
    }
    """

    def getFeaturesFromJSON(
        self, data, workspace: Path, img_path: Path, view: str
    ) -> list[dict]:
        """Get features from Pixano annotation file

        Args:
            data (str): raw json
            workspace (Path): Data path
            img_path (Path): Image path
            view (str): Image view name

        Raises:
            Exception: Image not found

        Returns:
            list[dict]: List of features
        """

        img_list = None
        if img_path is not None:
            self.info["images_path"] = str(workspace / img_path)
            img_list = sorted(list((workspace / img_path / view).glob("*.jpg")))
            if len(img_list) != len(data["data"]["children"]):
                print(
                    f"WARNING: Number ({len(img_list)}) of files in {img_path / view} differ \
                      than number {len(data['data']['children'])} of defined files in annotation file"
                )
                img_list = img_list[: len(data["data"]["children"])]
        else:
            self.info["images_path"] = str(workspace / data["data"]["path"])

        wraped_ds = []
        categories = set()
        # first we create Features with id, filename, width, height
        for idx, file_item in enumerate(tqdm(data["data"]["children"])):
            if img_list is not None:
                image = Image.open(img_list[idx])
                image_filename = str(Path(view) / Path(img_list[idx]).name)
            else:
                # Old Pixano format can be surprising... We need to check how to build file path
                fpath = workspace / data["data"]["path"] / file_item["path"]
                if not fpath.is_file():
                    fpath = (
                        workspace / data["data"]["path"] / Path(file_item["path"]).name
                    )
                    image_filename = fpath.name  # get rid of path in filename
                    if not fpath.is_file():
                        raise Exception("Unable to find image:" + str(fpath))
                image = Image.open(fpath)

            # keep real image size before thumbnailing
            img_w = image.width
            img_h = image.height

            image.thumbnail((128, 128))
            image_thumb = transforms.image_to_binary(image)

            feats = {
                "id": str(file_item["timestamp"]),
                view: {
                    "uri": image_filename,
                    "bytes": None,
                    "preview_bytes": image_thumb,
                },
                view + ".width": img_w,
                view + ".height": img_h,
                "objects": [],
            }
            wraped_ds.append(feats)

        # then we put annotations inside
        for ann in data["annotations"]:
            lfeats = [f for f in wraped_ds if f["id"] == str(ann["timestamp"])]
            feats = lfeats[0]
            w = feats[view + ".width"]
            h = feats[view + ".height"]
            categories.add(ann["category"])
            rle = None
            bbox = [0] * 4  # None  -- should be ok with None, but it's not
            if "geometry" in ann:
                if (
                    ann["geometry"]["type"] == "polygon" and ann["geometry"]["vertices"]
                ):  # Polygon
                    # we have normalized coords, we must denorm before making RLE
                    if not isnan(ann["geometry"]["vertices"][0]):
                        if len(ann["geometry"]["vertices"]) > 4:
                            denorm = transforms.denormalize(
                                ann["geometry"]["vertices"], w, h
                            )
                            rles = mask_api.frPyObjects([denorm], h, w)
                            rle = mask_api.merge(rles)
                        else:
                            print(
                                "Polygon with 2 or less points. Discarded\n",
                                ann["geometry"],
                            )
                elif (
                    ann["geometry"]["type"] == "mpolygon"
                    and ann["geometry"]["mvertices"]
                ):  # MultiPolygon
                    if not isnan(ann["geometry"]["mvertices"][0][0]):
                        denorm = [
                            transforms.denormalize(poly, w, h)
                            for poly in ann["geometry"]["mvertices"]
                        ]
                        rles = mask_api.frPyObjects(denorm, h, w)
                        rle = mask_api.merge(rles)
                elif (
                    ann["geometry"]["type"] == "rectangle"
                    and ann["geometry"]["vertices"]
                ):  # BBox
                    if not isnan(ann["geometry"]["vertices"][0]):
                        denorm = transforms.denormalize(
                            [ann["geometry"]["vertices"]], w, h
                        )
                        bbox = transforms.xyxy_to_xywh(denorm)
                elif (
                    ann["geometry"]["type"] == "graph" and ann["geometry"]["vertices"]
                ):  # Keypoints
                    print("Keypoints are not implemented yet")
                else:
                    # print('Unknown geometry', ann['geometry']['type'])  # log can be annoying if many...
                    pass
            else:
                print("No geometry?")  # Ca peut etre un mask, ou 3d, trackink... etc.

            # TMP to check
            if isinstance(rle, list) and len(rle) > 1:
                print("WARNING - MULTI RLE SPOTTED !!", len(rle), rle)

            feats["objects"].append(
                arrow_types.ObjectAnnotation(
                    id=str(ann["id"]),
                    view_id=view,
                    is_group_of=False,
                    # category_id=  # pas de categoryId, on le crée ensuite
                    category_name=ann["category"],
                    bbox=bbox,
                    mask=rle,
                ).dict()
            )

        # create category id based on existing categories index
        ordered_cat = list(categories)
        ordered_cat.sort()
        dict_cat_idx = dict([(cat, idx) for idx, cat in enumerate(ordered_cat)])
        for feats in wraped_ds:
            for obj in feats["objects"]:
                obj["category_id"] = dict_cat_idx[obj["category_name"]]
        return wraped_ds

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iter_data)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import base64
from typing import IO

import pyarrow as pa
from etils import epath


class Image:
    """Image type using URI string or bytes"""

    def __init__(
        self,
        uri: str,
        bytes: bytes,
        preview_bytes: bytes,
        uri_prefix: epath.PathLike = None,
    ):
        """Creates image from UIR, bytes and preview

        Attributes:
            uri (str): Image URI
            bytes (bytes): Image bytes
            preview_bytes (bytes): Image preview bytes
            uri_prefix (epath.PathLike, optional): Image URI prefix. Defaults to None.
        """
        self._uri = uri
        self._bytes = bytes
        self._preview_bytes = preview_bytes

        self.uri_prefix = uri_prefix

    @property
    def bytes(self) -> bytes:
        if self._bytes is not None:
            return self._bytes
        elif self._uri is not None:
            print(f"loading {self._uri}")
            with self.open() as f:
                return f.read()
        else:
            return None

    @property
    def preview_url(self) -> str:
        encoded = base64.b64encode(self._preview_bytes).decode("utf-8")
        url = f"data:image;base64,{encoded}"
        return url

    @property
    def url(self) -> str:
        # TODO need to check if not None
        data = self.bytes
        if data is not None:
            encoded = base64.b64encode(data).decode("utf-8")
            url = f"data:image;base64,{encoded}"
            return url
        else:
            return ""

    # TODO add prefix/auth/http/s3 ...
    def open(self) -> IO:
        if self.uri_prefix is not None:
            uri = self.uri_prefix / self._uri  # type: ignore
        else:
            uri = self._uri
        return open(uri, "rb")

    def display(self, preview=False):
        from IPython.core.display import Image as IPyImage

        if preview:
            data = self._preview_bytes
        else:
            data = self._bytes

        inferred_format = IPyImage(data).format
        encoded = base64.b64encode(data).decode("utf-8")
        url = f"data:image;base64,{encoded}"
        return IPyImage(url=url, format=inferred_format)


class ImageType(pa.ExtensionType):
    """Externalized image type containing the URI string in UTF-8"""

    def __init__(self):
        super(ImageType, self).__init__(
            pa.struct(
                [
                    pa.field("uri", pa.utf8()),
                    pa.field("bytes", pa.binary()),
                    pa.field("preview_bytes", pa.binary()),
                ]
            ),
            "Image",
        )

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return ImageType()

    def __arrow_ext_scalar_class__(self):
        return ImageScalar


class ImageScalar(pa.ExtensionScalar):
    def as_py(self) -> Image:
        return Image(
            self.value["uri"].as_py(),
            self.value["bytes"].as_py(),
            self.value["preview_bytes"].as_py(),
        )


class CompressedRLEType(pa.ExtensionType):
    """Segmentation mask type as PyArrow StructType"""

    def __init__(self):
        super(CompressedRLEType, self).__init__(
            pa.struct(
                [
                    pa.field("size", pa.list_(pa.int32())),
                    pa.field("counts", pa.binary()),
                ]
            ),
            "mask[rle]",
        )

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return CompressedRLEType()

    def __arrow_ext_serialize__(self):
        return b""


def is_image_type(t: pa.DataType) -> bool:
    """Returns True if value is an instance of ImageType

    Args:
        t (pa.DataType): Value to check

    Returns:
        bool: Type checking response
    """
    return isinstance(t, ImageType)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from .boxes import denormalize, mask_to_bbox, normalize, xywh_to_xyxy, xyxy_to_xywh
from .image import (
    binary_to_base64,
    depth_array_to_gray,
    depth_file_to_binary,
    image_to_binary,
    mask_to_polygons,
    mask_to_rle,
    rle_to_mask,
    rle_to_polygons,
    rle_to_urle,
)
from .labels import coco_names_80, coco_names_91, voc_names

__all__ = [
    "normalize",
    "denormalize",
    "mask_to_bbox",
    "xywh_to_xyxy",
    "xyxy_to_xywh",
    "image_to_binary",
    "binary_to_base64",
    "depth_file_to_binary",
    "depth_array_to_gray",
    "mask_to_rle",
    "rle_to_mask",
    "rle_to_polygons",
    "mask_to_polygons",
    "rle_to_urle",
    "coco_names_80",
    "coco_names_91",
    "voc_names",
]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import base64
from io import BytesIO
from itertools import groupby

import cv2
import numpy as np
from PIL import Image
from pycocotools import mask as mask_api


def image_to_binary(image: Image.Image, format: str = "PNG") -> bytes:
    """Encode image from Pillow to binary

    Args:
        image (Image.Image): Image as Pillow
        format (str, optional): Image file extension. Defaults to "PNG".

    Returns:
        bytes: Image as binary
    """

    with BytesIO() as output_bytes:
        image.save(output_bytes, format)
        binary = output_bytes.getvalue()

    return binary


def binary_to_base64(binary: bytes) -> str:
    """Encode image from binary to base 64

    Args:
        binary (bytes): Image as binary

    Returns:
        str: Image as base 64
    """

    b64 = base64.b64encode(binary).decode("utf-8")
    return f"data:image/png;base64,{b64}"


def depth_file_to_binary(depth_path: str) -> bytes:
    """Encode depth file to RGB image in binary

    Args:
        depth_path (str): Depth file path

    Returns:
        bytes: Depth file as RGB image in binary
    """

    depth = cv2.imread(depth_path, cv2.IMREAD_ANYDEPTH).astype(np.float32)
    depth = depth_array_to_gray(depth)
    depth_rgb = Image.fromarray(depth)

    return image_to_binary(depth_rgb)


def depth_array_to_gray(
    depth: np.ndarray,
    valid_start: float = 0.2,
    valid_end: float = 1,
    scale: float = 1.0,
) -> np.ndarray:
    """Encode depth array to gray levels

    Args:
        depth (np.ndarray): Depth array
        valid_start (float, optional): Valid start. Defaults to 0.2.
        valid_end (float, optional): Valid end. Defaults to 1.
        scale (float, optional): Scale. Defaults to 1.0.

    Returns:
        np.ndarray: Depth array in gray levels
    """

    mask = depth > 1

    # Scale gives depth in mm
    depth_n = depth * scale * 0.001

    if mask.sum() > 0:
        depth_n[mask] -= depth_n[mask].min()
        depth_n[mask] /= depth_n[mask].max() / (valid_end - valid_start)
        depth_n[mask] += valid_start

    depth_n *= 255
    depth_n = cv2.applyColorMap(
        depth_n[:, :, np.newaxis].astype(np.uint8), cv2.COLORMAP_PLASMA
    )

    return depth_n


def mask_to_rle(mask: Image.Image) -> dict:
    """Encode mask from Pillow or NumPy array to RLE

    Args:
        mask (Image.Image): Mask as Pillow or NumPy array

    Returns:
        dict: Mask as RLE
    """

    mask_array = np.asfortranarray(mask)
    return mask_api.encode(mask_array)


def rle_to_mask(rle: dict) -> np.ndarray:
    """Decode mask from RLE to NumPy array

    Args:
        rle (dict): Mask as RLE

    Returns:
        np.ndarray: Mask as NumPy array
    """

    return mask_api.decode(rle)


def rle_to_polygons(rle: dict) -> list[list]:
    """Encode mask from RLE to polygons

    Args:
        rle (dict): Mask as RLE

    Returns:
        list[list]: Mask as polygons
    """

    # If mask is empty or in wrong format
    if not rle or "size" not in rle:
        return []

    h, w = rle["size"]
    polygons, _ = mask_to_polygons(rle_to_mask(rle))

    # Normalize point coordinates
    for p in polygons:
        p[::2] /= w
        p[1::2] /= h

    # Cast to python list
    polygons = [p.tolist() for p in polygons]

    return polygons


def mask_to_polygons(mask: np.ndarray) -> tuple[list, bool]:
    """Encode mask from NumPy array to polygons

    Args:
        mask (np.ndarray): Mask as NumPy array

    Returns:
        list: Mask as polygons
        bool: Mask has holes
    """

    # Some versions of cv2 does not support incontiguous arr
    mask = np.ascontiguousarray(mask)

    # cv2.RETR_CCOMP flag retrieves all the contours and arranges them to a 2-level hierarchy.
    # External contours (boundary) of the object are placed in hierarchy-1.
    # Internal contours (holes) are placed in hierarchy-2.
    # cv2.CHAIN_APPROX_NONE flag gets vertices of polygons from contours.
    res = cv2.findContours(mask.astype("uint8"), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)
    hierarchy = res[-1]

    # If mask is empty
    if hierarchy is None:
        return [], False

    # Check if mask has holes
    has_holes = (hierarchy.reshape(-1, 4)[:, 3] >= 0).sum() > 0

    res = res[-2]
    res = [x.flatten() for x in res]

    # The coordinates from OpenCV are integers in range [0, W-1 or H-1].
    # We add 0.5 to turn them into real-value coordinate space. A better solution
    # would be to first +0.5 and then dilate the returned polygon by 0.5.
    res = [x + 0.5 for x in res if len(x) >= 6]

    return res, has_holes


def rle_to_urle(rle: dict) -> dict:
    """Encode mask from RLE to uncompressed RLE

    Args:
        rle (dict): Mask as RLE

    Returns:
        dict: Mask as uncompressed RLE
    """

    mask = rle_to_mask(rle)
    urle = {"counts": [], "size": list(mask.shape)}
    counts = urle.get("counts")

    for i, (value, elements) in enumerate(groupby(mask.ravel(order="F"))):
        if i == 0 and value == 1:
            counts.append(0)
        counts.append(len(list(elements)))

    return urle



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Any, Optional

from pydantic import BaseModel


class Point2d(BaseModel):
    """Point2d

    Attributes:
        x (float): x coordinate
        y (float): y coordinate
    """

    x: float
    y: float


class Rectangle(BaseModel):
    """Rectangle

    Attributes:
        bottom_left (Point2d): bottom-left corner of the rectangle
        width (float): width of the rectangle
        height (float): height of the rectangle
    """

    bottom_left: Point2d
    width: float
    height: float


class Polygon(BaseModel):
    """Polygon

    Attributes:
        points (list[Point2d]): Polygon points list
    """

    points: list[Point2d]


MultiPolygon = list[Polygon]


class Point3d(BaseModel):
    """Point3d structure

    Attributes:
        x (float): x coordinate
        y (float): y coordinate
        z (float): z coordinate
    """

    x: float
    y: float
    z: float


class Mask(BaseModel):
    pass


class RLE(BaseModel):
    """RLE Mask (Run Length Encoding)

    Attributes:
        size (list[int]): Image size (width, height)
        counts (bytes): RLE Mask
    """

    size: list[int]
    counts: bytes


class Category(BaseModel):
    """Category

    Attributes:
        name (str): Category name
        id (int, optional): Category ID
    """

    name: str
    id: Optional[int]


class ObjectAnnotation(BaseModel):
    """ObjectAnnotation

    Attributes:
        id (str): Annotation unique ID
        view_id (str, optional): View ID (e.g. 'image', 'cam_2')
        bbox (list[float], optional): Bounding box coordinates in xywh format (using top left point as reference)
        bbox_source (str, optional): Bounding box source
        bbox_confidence (float, optional): Bounding box confidence
        is_group_of (bool, optional): is_group_of
        is_difficult (bool, optional): is_difficult
        is_truncated (bool, optional): is_truncated
        mask (RLE, optional): Mask
        mask_source (str, optional): Mask source
        area (float, optional): area
        pose (dict[str, list[float]], optional): Pose
        category_id (int, optional): Category ID
        category_name (str, optional): Category name
        identity (str, optional): Identity
    """

    id: str
    view_id: Optional[str] = None
    bbox: Optional[list[float]] = None
    bbox_source: Optional[str] = None
    bbox_confidence: Optional[float] = None
    is_group_of: Optional[bool] = None
    is_difficult: Optional[bool] = None
    is_truncated: Optional[bool] = None
    mask: Optional[RLE] = None
    mask_source: Optional[str] = None
    area: Optional[float] = None
    pose: Optional[dict[str, list[float]]] = None
    category_id: Optional[int] = None
    category_name: Optional[str] = None
    identity: Optional[str] = None


class Feature(BaseModel):
    """Feature

    Attributes:
        name (str): Feature name
        dtype (str): Feature dtype
        value (Any): Feature value
    """

    name: str
    dtype: str
    value: Any


Features = list[Feature]


class DatasetSpec(BaseModel):
    """DatasetSpec

    Attributes:
        id (str): Dataset ID
        name (str): Dataset name
        description (str): Dataset description
        num_elements (int): Number of elements in dataset
        preview (str): Dataset preview
        splits (dict[str, int], optional): Dataset splits
        category_index (dict[str, int], optional): Dataset category index
    """

    id: str
    name: str
    description: str
    num_elements: int
    preview: str
    splits: Optional[dict[str, int]] = None
    category_index: Optional[dict[str, int]] = None



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path

import click
import fastapi
import pkg_resources
import uvicorn
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from .main import app, settings

logo = """
                             ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▒
                           ▒▓▒                                       ▓
              █▓          ▓▓                                         ▓
                        ▒▓▒   ▒▓▓▓▓▓ ▒▓   ▓▒▒▓▓▓▓▒     ▒▓▓▓▓▓▒       ▓
▓▓ ▓▓▓▓▓▓▒    ▓▒  ▒▓   ▓▒   ▒▓      ▓██   ██     ▓▒  ▒█▓     ▒█▓     ▓
███      ▓▓   █▒   ▓█      ▒█▒       ██   █▒     ▒█  ▓▒        █     ▓
██        █▒  █▒    ▒█▓     █▒       ██   █▒     ▒█  ▓▓        █     ▓
██        █▒  █▒    ▓█▓█     █▒     ▓██   █▒     ▒█   ██     ▓█▒     ▓
██▓▒    ▒▓▒   █▒   ▓▒  ▒█▓    ▒▓▓▓▓▓ ▒▓   ▓      ▒▓    ▒▓▓▓▓▓▒       ▓
██  ▒▓▓▒▒     ▒   ▒      ▒▓▒                                         ▓
██                         █▓                                        ▓
"""

ASSETS_PATH = pkg_resources.resource_filename("pixano", "apps/explorer/dist/assets")
TEMPLATE_PATH = pkg_resources.resource_filename("pixano", "apps/explorer/dist")


@click.command(context_settings={"auto_envvar_prefix": "UVICORN"})
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="Bind socket to this port.",
    show_default=True,
)
@click.argument("library_dir")
def main(host: str, port: int, library_dir: str):
    templates = Jinja2Templates(directory=TEMPLATE_PATH)

    settings.data_dir = Path(library_dir)

    @app.get("/", response_class=HTMLResponse)
    def main(request: fastapi.Request):
        return templates.TemplateResponse("index.html", {"request": request})

    app.mount("/assets", StaticFiles(directory=ASSETS_PATH), name="assets")
    uvicorn.run(app, host=host, port=port)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import os
from abc import abstractmethod
from io import BytesIO
from pathlib import Path

import cv2
import duckdb
import numpy as np
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from onnxruntime import InferenceSession
from tqdm.auto import tqdm

from pixano.transforms import binary_to_base64

from .inference_model import InferenceModel


class OnlineModel(InferenceModel):
    """InferenceModel class for online interactive annotation

    Attributes:
        name (str): Model name
        id (str): Model ID
        device (str): Model GPU or CPU device
        source (str): Model source
        info (str): Additional model info
        onnx_path (Path): ONNX Model Path
        onnx_session (onnxruntime.InferenceSession): ONNX session
        working (dict): Dictionary of current working data
    """

    def __init__(
        self,
        name: str,
        id: str = "",
        device: str = "",
        source: str = "",
        info: str = "",
    ) -> None:
        super().__init__(name, id, device, source, info)
        self.working = {}

    @abstractmethod
    def __call__(self, input: dict[str, np.ndarray]) -> np.ndarray:
        """Return model annotation based on user input

        Args:
            input (dict[str, np.ndarray]): User input

        Returns:
            np.ndarray: Model annotation masks
        """

        pass

    @abstractmethod
    def process_batch(
        self,
        batch: pa.RecordBatch,
        view: str,
        media_dir: Path,
    ) -> list[np.ndarray]:
        """Precompute embeddings for a batch

        Args:
            batch (pa.RecordBatch): Input batch
            view (str): Dataset view
            media_dir (Path): Media location

        Returns:
            list[np.ndarray]: Model embeddings as NumPy arrays
        """

        pass

    def process_dataset(
        self,
        input_dir: Path,
        views: list[str],
        splits: list[str] = [],
        batch_size: int = 1,
    ) -> Path:
        """Precompute embeddings for a parquet dataset

        Args:
            input_dir (Path): Input parquet location
            views (list[str]): Dataset views
            splits (list[str], optional): Dataset splits, all if []. Defaults to [].
            batch_size (int, optional): Rows per batch. Defaults to 1.

        Returns:
            Path: Output parquet location
        """

        output_dir = input_dir / f"db_embed_{self.id}"

        # Load spec.json
        with open(input_dir / "spec.json", "r") as f:
            spec_json = json.load(f)

        # If no splits given, select all splits
        if splits == []:
            splits = [s.name for s in os.scandir(input_dir / "db") if s.is_dir()]

        # Create schema
        fields = [pa.field("id", pa.string())]
        fields.extend([pa.field(f"{view}_embedding", pa.binary()) for view in views])
        schema = pa.schema(fields)

        # Iterate on splits
        for split in splits:
            # List dataset files
            files = sorted((input_dir / "db" / split).glob("*.parquet"))

            # Create folder
            split_dir = output_dir / split
            split_dir.mkdir(parents=True, exist_ok=True)

            # Check for already processed files
            processed = [p.name for p in split_dir.glob("*.parquet")]

            # Iterate on files
            for file in tqdm(files, desc=split, position=0):
                # Process only remaining files
                if file.name not in processed:
                    # Load file into batches
                    table = pq.read_table(file)
                    batches = table.to_batches(max_chunksize=batch_size)

                    # Iterate on batches
                    data = {field.name: [] for field in schema}
                    for batch in tqdm(batches, position=1, desc=file.name):
                        # Add row IDs
                        data["id"].extend([str(row) for row in batch["id"]])
                        # Iterate on views
                        for view in views:
                            view_emb = self.process_batch(
                                batch, view, input_dir / "media"
                            )
                            for emb in view_emb:
                                emb_bytes = BytesIO()
                                np.save(emb_bytes, emb)
                                data[f"{view}_embedding"].append(emb_bytes.getvalue())

                    # Save to file
                    pq.write_table(
                        pa.Table.from_pydict(data, schema=schema),
                        split_dir / file.name,
                    )

                    # Save.json
                    super().save_json(
                        output_dir=output_dir,
                        filename="embed",
                        spec_json=spec_json,
                        num_elements=pq.read_metadata(split_dir / file.name).num_rows,
                    )

        return output_dir

    @abstractmethod
    def export_onnx_model(self) -> Path:
        """Export Torch model to ONNX

        Returns:
            Path: ONNX model path
        """

        pass

    def set_onnx_model(self, onnx_path: Path = None):
        """Set existing ONNX model path or export model to ONNX

        Args:
            onnx_path (Path, optional): ONNX Model path. Defaults to None.
        """

        if onnx_path == None:
            onnx_path = self.export_onnx_model()

        self.onnx_path = Path(onnx_path)

    def set_dataset(self, input_dir: Path):
        """Set current working dataset

        Args:
            input_dir (Path): Dataset path
        """

        self.working["input_dir"] = Path(input_dir)
        self.working["embed_dir"] = Path(input_dir) / f"db_embed_{self.id}"

    def set_image(self, id: str, view: str) -> str:
        """Set current working image and return embedding

        Args:
            id (str): Row ID
            view (str): View ID

        Returns:
            str: Image embedding in base 64
        """

        # Load datasets
        split = ds.partitioning(pa.schema([("split", pa.string())]), flavor="hive")
        input_ds = ds.dataset(
            self.working["input_dir"] / "db",
            partitioning=split,
        )
        embed_ds = ds.dataset(
            self.working["embed_dir"],
            partitioning=split,
            ignore_prefixes=["info", "embed.json"],
        )

        # Set working image
        input_row = duckdb.query(f"SELECT * FROM input_ds WHERE id={id}").arrow()
        image = cv2.imread(
            str(self.working["input_dir"] / "media" / input_row[view][0].as_py()["uri"])
        )
        self.working["image"] = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Get embedding
        embed_row = duckdb.query(f"SELECT * FROM embed_ds WHERE id={id}").arrow()
        embedding = binary_to_base64(embed_row["image_embedding"][0].as_py())

        # Reshape and return embedding
        return embedding

    def open_onnx_session(self):
        """Open an ONNX session for interactive annotation"""

        self.onnx_session = InferenceSession(self.onnx_path.as_posix())

    def close_onnx_session(self):
        """Close the ONNX session for interactive annotation"""

        del self.onnx_session



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import pyarrow as pa

from .features import BBoxType, ObjectAnnotation, ObjectAnnotationType
from .image import CompressedRLEType, Image, ImageType, is_image_type

__all__ = [
    "BBoxType",
    "ObjectAnnotation",
    "ObjectAnnotationType",
    "CompressedRLEType",
    "Image",
    "ImageType",
    "is_image_type",
]


def convert_field(
    field_name: str, field_type: pa.DataType, field_data: list
) -> pa.Array:
    """Convert PyArrow ExtensionTypes properly

    Args:
        field_name (str): Name
        field_type (pa.DataType): Target PyArrow format
        field_data (list): Data in Python format

    Returns:
        pa.Array: Data in target PyArrow format
    """

    # If target format is an ExtensionType
    if isinstance(field_type, pa.ExtensionType):
        storage = pa.array(field_data, type=field_type.storage_type)
        return pa.ExtensionArray.from_storage(field_type, storage)

    # If target format is an extension of ListType
    elif pa.types.is_list(field_type):
        native_arr = pa.array(field_data)
        if isinstance(native_arr, pa.NullArray):
            return pa.nulls(len(native_arr), field_type)
        offsets = native_arr.offsets
        values = native_arr.values.to_numpy(zero_copy_only=False)
        return pa.ListArray.from_arrays(
            offsets,
            convert_field(f"{field_name}.elements", field_type.value_type, values),
        )

    # If target format is an extension of StructType
    elif pa.types.is_struct(field_type):
        native_arr = pa.array(field_data)
        if isinstance(native_arr, pa.NullArray):
            return pa.nulls(len(native_arr), field_type)
        arrays = []
        for subfield in field_type:
            sub_arr = native_arr.field(subfield.name)
            converted = convert_field(
                f"{field_name}.{subfield.name}",
                subfield.type,
                sub_arr.to_numpy(zero_copy_only=False),
            )
            arrays.append(converted)
        return pa.StructArray.from_arrays(arrays, fields=field_type)

    # For other target formats
    else:
        return pa.array(field_data, type=field_type)


def register_extension_types():
    """Register PyArrow ExtensionTypes"""

    types = [
        BBoxType(),
        CompressedRLEType(),
        ImageType(),
    ]

    for t in types:
        # Register ExtensionType
        try:
            pa.register_extension_type(t)
        # If ExtensionType is already registered
        except pa.ArrowKeyError:
            pass


def is_number(t: pa.DataType) -> bool:
    """Check if DataType is a a number (integer or float)

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is an integer or a float
    """

    return pa.types.is_integer(t) or pa.types.is_floating(t)


register_extension_types()



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
import os
from abc import abstractmethod
from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as pq
from tqdm.auto import tqdm

from pixano.core import arrow_types

from .inference_model import InferenceModel


class OfflineModel(InferenceModel):
    """InferenceModel class for offline pre-annotation

    Attributes:
        name (str): Model name
        id (str): Model ID
        device (str): Model GPU or CPU device
        source (str): Model source
        info (str): Additional model info
    """

    @abstractmethod
    def __call__(
        self,
        batch: pa.RecordBatch,
        view: str,
        media_dir: Path,
        threshold: float = 0.0,
    ) -> list[list[arrow_types.ObjectAnnotation]]:
        """Process batch

        Args:
            batch (pa.RecordBatch): Input batch
            view (str): Dataset view
            media_dir (Path): Media location
            threshold (float, optional): Confidence threshold. Defaults to 0.0.

        Returns:
            list[list[arrow_types.ObjectAnnotation]]: Model inferences as lists of ObjectAnnotation
        """

        pass

    def process_dataset(
        self,
        input_dir: Path,
        views: list[str],
        splits: list[str] = [],
        batch_size: int = 1,
        threshold: float = 0.0,
    ) -> Path:
        """Generate inferences for a parquet dataset

        Args:
            input_dir (Path): Input parquet location
            views (list[str]): Dataset views
            splits (list[str], optional): Dataset splits, all if []. Defaults to [].
            batch_size (int, optional): Rows per batch. Defaults to 1.
            threshold (float, optional): Confidence threshold for model predictions. Defaults to 0.0.

        Returns:
            Path: Output parquet location
        """

        output_dir = input_dir / f"db_infer_{self.id}"

        # Load spec.json
        with open(input_dir / "spec.json", "r") as f:
            spec_json = json.load(f)

        # If no splits given, select all splits
        if splits == []:
            splits = [s.name for s in os.scandir(input_dir / "db") if s.is_dir()]

        # Create schema
        schema = pa.schema(
            [
                pa.field("id", pa.string()),
                pa.field("objects", pa.list_(arrow_types.ObjectAnnotationType())),
            ]
        )

        # Iterate on splits
        for split in splits:
            # List dataset files
            files = sorted((input_dir / "db" / split).glob("*.parquet"))

            # Create folder
            split_dir = output_dir / split
            split_dir.mkdir(parents=True, exist_ok=True)

            # Check for already processed files
            processed = [p.name for p in split_dir.glob("*.parquet")]

            # Iterate on files
            for file in tqdm(files, desc=split, position=0):
                # Process only remaining files
                if file.name not in processed:
                    # Load file into batches
                    table = pq.read_table(file)
                    batches = table.to_batches(max_chunksize=batch_size)

                    # Iterate on batches
                    data = {field.name: [] for field in schema}
                    for batch in tqdm(batches, position=1, desc=file.name):
                        # Add row IDs
                        data["id"].extend([str(row) for row in batch["id"]])
                        # Iterate on views
                        batch_inf = []
                        for view in views:
                            batch_inf.append(
                                self(batch, view, input_dir / "media", threshold)
                            )
                        # Regroup view inferences by row
                        data["objects"].append(
                            [
                                inf.dict()
                                for row in range(len(batch["id"]))
                                for view_inf in batch_inf
                                for inf in view_inf[row]
                            ]
                        )

                    # Convert ExtensionTypes
                    arrays = []
                    for field_name, field_data in data.items():
                        arrays.append(
                            arrow_types.convert_field(
                                field_name=field_name,
                                field_type=schema.field(field_name).type,
                                field_data=field_data,
                            )
                        )

                    # Save to file
                    pq.write_table(
                        pa.Table.from_arrays(arrays, schema=schema),
                        split_dir / file.name,
                    )

                    # Save.json
                    super().save_json(
                        output_dir=output_dir,
                        filename="infer",
                        spec_json=spec_json,
                        num_elements=pq.read_metadata(split_dir / file.name).num_rows,
                    )

        return output_dir



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
#
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from pathlib import Path

from fastapi import Depends, FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi_pagination import Page, Params
from fastapi_pagination.api import add_pagination
from pydantic import BaseSettings

from pixano.core import arrow_types
from pixano.core.dataset import Dataset, DatasetInfo, InferenceDataset
from pixano.data import models

from . import db_utils


class Settings(BaseSettings):
    """Dataset library settings

    Attributes:
        data_dir (Path): Dataset library path
    """

    data_dir: Path = Path.cwd() / "library"


def find_dataset(dataset_id: str, settings: Settings) -> dict:
    """Return dataset path and info based on its ID

    Args:
        dataset_id (str): Dataset ID
        settings (Settings): Dataset library

    Returns:
        dict: Dataset path and info
    """

    for info_file in settings.data_dir.glob("*/spec.json"):
        print(info_file)
        info = DatasetInfo.parse_file(info_file)
        if info.id == dataset_id:
            return {"path": info_file.parent, "info": info}
    return None


def load_library(settings: Settings) -> list[DatasetInfo]:
    """Load all dataset info files in library

    Args:
        settings (Settings): Dataset library

    Returns:
        list[DatasetInfo]: Dataset info files
    """

    infos = []
    for info_file in sorted(list(settings.data_dir.glob("*/spec.json"))):
        print(info_file)
        info = DatasetInfo.parse_file(info_file)
        preview = info_file.parent / "preview.png"
        if preview.is_file():
            im = arrow_types.Image(
                uri="preview.png",
                bytes=None,
                preview_bytes=None,
                uri_prefix=info_file.parent,
            )
            info.preview = im.url
        infos.append(info)
    return infos


def load_dataset(dataset_id: str, settings: Settings) -> Dataset:
    """Load dataset based on its ID

    Args:
        dataset_id (str): Dataset ID
        settings (Settings): Dataset library

    Returns:
        Dataset: Dataset
    """

    ds = find_dataset(dataset_id, settings)
    if ds is not None:
        ds = Dataset(ds["path"])
    return ds


def load_dataset_stats(dataset_id: str, settings: Settings) -> dict:
    """Load dataset stats based on its ID

    Args:
        dataset_id (str): Dataset ID
        settings (Settings): Dataset Library

    Returns:
        dict: Dataset stats
    """

    stats = []
    ds = find_dataset(dataset_id, settings)
    if ds:
        stat_file = settings.data_dir / ds["path"].name / "db_feature_statistics.json"
        if stat_file.is_file():
            with open(stat_file, "r") as f:
                stats = json.load(f)
    return stats


def create_app(settings: Settings) -> FastAPI:
    """Run explorer app in library

    Args:
        settings (Settings): Dataset Library

    Raises:
        HTTPException: 404 error if dataset items are not found
        HTTPException: 404 error if dataset stats are not found
        HTTPException: 404 error if dataset item details are not found

    Returns:
        FastAPI: Explorer app
    """
    app = FastAPI()

    app.add_middleware(
        CORSMiddleware,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.get("/datasets", response_model=list[DatasetInfo])
    async def get_datasets_list():
        library = load_library(settings)
        print(library)
        return library

    @app.get("/datasets/{dataset_id}/items", response_model=Page[models.Features])
    async def get_dataset_items(dataset_id, params: Params = Depends()):
        dataset = load_dataset(dataset_id, settings)
        if dataset is None:
            raise HTTPException(status_code=404, detail="Item not found")
        ds = dataset.load()

        return db_utils.get_items(ds, params)

    @app.get("/datasets/{dataset_id}/stats")
    async def get_dataset_stats(
        dataset_id,
    ):
        stats = load_dataset_stats(dataset_id, settings)
        if stats is None:
            raise HTTPException(status_code=404, detail="Item not found")
        return stats

    @app.get("/datasets/{dataset_id}/items/{id}")
    async def get_dataset_item_details(dataset_id: str, id: str):
        dataset = load_dataset(dataset_id, settings)
        if dataset is None:
            raise HTTPException(status_code=404, detail="Item not found")

        ds = dataset.load()

        # load Inferences datasets
        tmp_ds = find_dataset(
            dataset_id, settings
        )  # ?? more clever way to get tmp_ds["path"].name ??
        infer_datasets = []
        if tmp_ds:
            for info_file in settings.data_dir.glob(
                tmp_ds["path"].name + "/db_*/infer.json"
            ):
                infer_datasets.append(InferenceDataset(info_file.parent).load())

        return db_utils.get_item_details(ds, id, dataset.media_path, infer_datasets)

    add_pagination(app)

    return app


settings = Settings()
app = create_app(settings)
